package java.lang;

import java.util.regex.Pattern;
import sun.misc.SharedSecrets;
import java.util.concurrent.TimeUnit;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import sun.security.action.GetPropertyAction;
import java.util.Locale;
import java.util.regex.Matcher;
import java.util.ArrayList;
import java.io.FileInputStream;
import java.util.Map;
import java.io.IOException;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.io.FileDescriptor;
import java.io.FileOutputStream;
import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import sun.misc.JavaIOFileDescriptorAccess;

final class ProcessImpl extends Process
{
    private static final JavaIOFileDescriptorAccess fdAccess;
    private static final int VERIFICATION_CMD_BAT = 0;
    private static final int VERIFICATION_WIN32 = 1;
    private static final int VERIFICATION_WIN32_SAFE = 2;
    private static final int VERIFICATION_LEGACY = 3;
    private static final char[][] ESCAPE_VERIFICATION;
    private static final char DOUBLEQUOTE = '\"';
    private static final char BACKSLASH = '\\';
    private long handle;
    private OutputStream stdin_stream;
    private InputStream stdout_stream;
    private InputStream stderr_stream;
    private static final int STILL_ACTIVE;
    
    private static FileOutputStream newFileOutputStream(File file, boolean b) throws IOException {
        if (b) {
            String path = file.getPath();
            SecurityManager securityManager = System.getSecurityManager();
            if (securityManager != null) {
                securityManager.checkWrite(path);
            }
            long openForAtomicAppend = openForAtomicAppend(path);
            final FileDescriptor fileDescriptor = new FileDescriptor();
            ProcessImpl.fdAccess.setHandle(fileDescriptor, openForAtomicAppend);
            return (FileOutputStream)AccessController.doPrivileged(new PrivilegedAction<FileOutputStream>() {
                @Override
                public FileOutputStream run() {
                    return new FileOutputStream(fileDescriptor);
                }
            });
        }
        else {
            return new FileOutputStream(file);
        }
    }
    
    static Process start(String[] array, Map<String, String> map, String s, ProcessBuilder.Redirect[] array2, boolean b) throws IOException {
        String environmentBlock = ProcessEnvironment.toEnvironmentBlock(map);
        FileInputStream fileInputStream = null;
        FileOutputStream fileOutputStream = null;
        FileOutputStream fileOutputStream2 = null;
        try {
            long[] array3;
            if (array2 == null) {
                array3 = new long[] { -1L, -1L, -1L };
            }
            else {
                array3 = new long[3];
                if (array2[0] == ProcessBuilder.Redirect.PIPE) {
                    array3[0] = -1L;
                }
                else if (array2[0] == ProcessBuilder.Redirect.INHERIT) {
                    array3[0] = ProcessImpl.fdAccess.getHandle(FileDescriptor.in);
                }
                else {
                    fileInputStream = new FileInputStream(array2[0].file());
                    array3[0] = ProcessImpl.fdAccess.getHandle(fileInputStream.getFD());
                }
                if (array2[1] == ProcessBuilder.Redirect.PIPE) {
                    array3[1] = -1L;
                }
                else if (array2[1] == ProcessBuilder.Redirect.INHERIT) {
                    array3[1] = ProcessImpl.fdAccess.getHandle(FileDescriptor.out);
                }
                else {
                    fileOutputStream = newFileOutputStream(array2[1].file(), array2[1].append());
                    array3[1] = ProcessImpl.fdAccess.getHandle(fileOutputStream.getFD());
                }
                if (array2[2] == ProcessBuilder.Redirect.PIPE) {
                    array3[2] = -1L;
                }
                else if (array2[2] == ProcessBuilder.Redirect.INHERIT) {
                    array3[2] = ProcessImpl.fdAccess.getHandle(FileDescriptor.err);
                }
                else {
                    fileOutputStream2 = newFileOutputStream(array2[2].file(), array2[2].append());
                    array3[2] = ProcessImpl.fdAccess.getHandle(fileOutputStream2.getFD());
                }
            }
            return new ProcessImpl(array, environmentBlock, s, array3, b);
        }
        finally {
            try {
                if (fileInputStream != null) {
                    fileInputStream.close();
                }
            }
            finally {
                try {
                    if (fileOutputStream != null) {
                        fileOutputStream.close();
                    }
                }
                finally {
                    if (fileOutputStream2 != null) {
                        fileOutputStream2.close();
                    }
                }
            }
        }
    }
    
    private static String[] getTokensFromCommand(String input) {
        ArrayList list = new ArrayList(8);
        Matcher matcher = ProcessImpl.LazyPattern.PATTERN.matcher(input);
        while (matcher.find()) {
            list.add(matcher.group());
        }
        return (String[])list.toArray(new String[list.size()]);
    }
    
    private static String createCommandLine(int n, String s, String[] array) {
        StringBuilder sb = new StringBuilder(80);
        sb.append(s);
        for (int i = 1; i < array.length; ++i) {
            sb.append(' ');
            String s2 = array[i];
            if (needsEscaping(n, s2)) {
                sb.append('\"');
                if (n == 2) {
                    for (int length = s2.length(), j = 0; j < length; ++j) {
                        int char1 = s2.charAt(j);
                        if (char1 == '\"') {
                            int countLeadingBackslash = countLeadingBackslash(n, s2, j);
                            while (countLeadingBackslash-- > 0) {
                                sb.append('\\');
                            }
                            sb.append('\\');
                        }
                        sb.append((char)char1);
                    }
                }
                else {
                    sb.append(s2);
                }
                int countLeadingBackslash2 = countLeadingBackslash(n, s2, s2.length());
                while (countLeadingBackslash2-- > 0) {
                    sb.append('\\');
                }
                sb.append('\"');
            }
            else {
                sb.append(s2);
            }
        }
        return sb.toString();
    }
    
    private static String unQuote(String s) {
        int length = s.length();
        return (length >= 2 && s.charAt(0) == '\"' && s.charAt(length - 1) == '\"') ? s.substring(1, length - 1) : s;
    }
    
    private static boolean needsEscaping(int n, String s) {
        String unQuote = unQuote(s);
        int n2 = !s.equals(unQuote);
        int n3 = unQuote.indexOf(34) >= 0;
        switch (n) {
            case 0: {
                if (!(n3 == 0)) {
                    throw new IllegalArgumentException("Argument has embedded quote, use the explicit CMD.EXE call.");
                }
                break;
            }
            case 2: {
                if (n2 && n3) {
                    throw new IllegalArgumentException("Malformed argument has embedded quote: " + unQuote);
                }
                break;
            }
        }
        if (n2 == 0) {
            char[] array = ProcessImpl.ESCAPE_VERIFICATION[n];
            int i = 0;
            while (i < array.length) {
                if (s.indexOf((int)array[i]) >= 0) {
                    return true;
                }
                ++i;
            }
        }
        return false;
    }
    
    private static String getExecutablePath(String s) throws IOException {
        String unQuote = unQuote(s);
        if (unQuote.indexOf(34) >= 0) {
            throw new IllegalArgumentException("Executable name has embedded quote, split the arguments: " + unQuote);
        }
        return new File(unQuote).getPath();
    }
    
    private boolean isExe(String pathname) {
        String upperCase = new File(pathname).getName().toUpperCase(Locale.ROOT);
        return upperCase.endsWith(".EXE") || upperCase.indexOf(46) < 0;
    }
    
    private boolean isShellFile(String s) {
        String upperCase = s.toUpperCase();
        return upperCase.endsWith(".CMD") || upperCase.endsWith(".BAT");
    }
    
    private String quoteString(String s) {
        return new StringBuilder(s.length() + 2).append('\"').append(s).append('\"').toString();
    }
    
    private static int countLeadingBackslash(int n, CharSequence charSequence, int n2) {
        if (n == 0) {
            return 0;
        }
        int n3;
        for (n3 = n2 - 1; n3 >= 0 && charSequence.charAt(n3) == '\\'; --n3) {}
        return n2 - 1 - n3;
    }
    
    private ProcessImpl(String[] tokensFromCommand, String s, String s2, final long[] array, boolean b) throws IOException {
        this.handle = 0L;
        SecurityManager securityManager = System.getSecurityManager();
        int n = !"false".equalsIgnoreCase(GetPropertyAction.privilegedGetProperty("jdk.lang.Process.allowAmbiguousCommands", (securityManager == null) ? "true" : "false"));
        String s4;
        if (n && securityManager == null) {
            String s5 = new File(tokensFromCommand[0]).getPath();
            if (needsEscaping(3, s5)) {
                s5 = this.quoteString(s5);
            }
            s4 = createCommandLine(3, s5, tokensFromCommand);
        }
        else {
            String s3;
            try {
                s3 = getExecutablePath(tokensFromCommand[0]);
            }
            catch (IllegalArgumentException ex) {
                StringBuilder sb = new StringBuilder();
                for (int length = tokensFromCommand.length, i = 0; i < length; ++i) {
                    sb.append(tokensFromCommand[i]).append(' ');
                }
                tokensFromCommand = getTokensFromCommand(sb.toString());
                s3 = getExecutablePath(tokensFromCommand[0]);
                if (securityManager != null) {
                    securityManager.checkExec(s3);
                }
            }
            s4 = createCommandLine((n ? this.isShellFile(s3) : (!this.isExe(s3))) ? 0 : (n ? 1 : 2), this.quoteString(s3), tokensFromCommand);
        }
        this.handle = create(s4, s, s2, array, b);
        AccessController.doPrivileged(new PrivilegedAction<Void>() {
            @Override
            public Void run() {
                if (array[0] == -1L) {
                    ProcessImpl.this.stdin_stream = ProcessBuilder.NullOutputStream.INSTANCE;
                }
                else {
                    FileDescriptor fdObj = new FileDescriptor();
                    ProcessImpl.fdAccess.setHandle(fdObj, array[0]);
                    ProcessImpl.this.stdin_stream = new BufferedOutputStream(new FileOutputStream(fdObj));
                }
                if (array[1] == -1L) {
                    ProcessImpl.this.stdout_stream = ProcessBuilder.NullInputStream.INSTANCE;
                }
                else {
                    FileDescriptor fdObj2 = new FileDescriptor();
                    ProcessImpl.fdAccess.setHandle(fdObj2, array[1]);
                    ProcessImpl.this.stdout_stream = new BufferedInputStream(new FileInputStream(fdObj2));
                }
                if (array[2] == -1L) {
                    ProcessImpl.this.stderr_stream = ProcessBuilder.NullInputStream.INSTANCE;
                }
                else {
                    FileDescriptor fdObj3 = new FileDescriptor();
                    ProcessImpl.fdAccess.setHandle(fdObj3, array[2]);
                    ProcessImpl.this.stderr_stream = new FileInputStream(fdObj3);
                }
                return null;
            }
        });
    }
    
    @Override
    public OutputStream getOutputStream() {
        return this.stdin_stream;
    }
    
    @Override
    public InputStream getInputStream() {
        return this.stdout_stream;
    }
    
    @Override
    public InputStream getErrorStream() {
        return this.stderr_stream;
    }
    
    @Override
    protected void finalize() {
        closeHandle(this.handle);
    }
    
    private static native int getStillActive();
    
    @Override
    public int exitValue() {
        int exitCodeProcess = getExitCodeProcess(this.handle);
        if (exitCodeProcess == ProcessImpl.STILL_ACTIVE) {
            throw new IllegalThreadStateException("process has not exited");
        }
        return exitCodeProcess;
    }
    
    private static native int getExitCodeProcess(long p0);
    
    @Override
    public int waitFor() throws InterruptedException {
        waitForInterruptibly(this.handle);
        if (Thread.interrupted()) {
            throw new InterruptedException();
        }
        return this.exitValue();
    }
    
    private static native void waitForInterruptibly(long p0);
    
    @Override
    public boolean waitFor(long duration, TimeUnit timeUnit) throws InterruptedException {
        long nanos = timeUnit.toNanos(duration);
        if (getExitCodeProcess(this.handle) != ProcessImpl.STILL_ACTIVE) {
            return true;
        }
        if (duration <= 0L) {
            return false;
        }
        long n = System.nanoTime() + nanos;
        do {
            long millis = TimeUnit.NANOSECONDS.toMillis(nanos + 999999L);
            if (millis < 0L) {
                millis = 2147483647L;
            }
            waitForTimeoutInterruptibly(this.handle, millis);
            if (Thread.interrupted()) {
                throw new InterruptedException();
            }
            if (getExitCodeProcess(this.handle) != ProcessImpl.STILL_ACTIVE) {
                return true;
            }
            nanos = n - System.nanoTime();
        } while (nanos > 0L);
        return getExitCodeProcess(this.handle) != ProcessImpl.STILL_ACTIVE;
    }
    
    private static native void waitForTimeoutInterruptibly(long p0, long p1);
    
    @Override
    public void destroy() {
        terminateProcess(this.handle);
    }
    
    @Override
    public Process destroyForcibly() {
        this.destroy();
        return this;
    }
    
    private static native void terminateProcess(long p0);
    
    @Override
    public boolean isAlive() {
        return isProcessAlive(this.handle);
    }
    
    private static native boolean isProcessAlive(long p0);
    
    private static synchronized native long create(String p0, String p1, String p2, long[] p3, boolean p4) throws IOException;
    
    private static native long openForAtomicAppend(String p0) throws IOException;
    
    private static native boolean closeHandle(long p0);
    
    static {
        fdAccess = SharedSecrets.getJavaIOFileDescriptorAccess();
        ESCAPE_VERIFICATION = new char[][] { { ' ', '\t', '<', '>', '&', '|', '^' }, { ' ', '\t', '<', '>' }, { ' ', '\t', '<', '>' }, { ' ', '\t' } };
        STILL_ACTIVE = getStillActive();
    }
    
    private static class LazyPattern
    {
        private static final Pattern PATTERN;
        
        static {
            PATTERN = Pattern.compile("[^\\s\"]+|\"[^\"]*\"");
        }
    }
}
